"""
The full documentation is avaiable here: https://docs.optimeering.no/

These are the functions available when importing __archimedes__.

Example usage:
>>> archimedes.list_ids()
...prints list of available time series
"""

from archimedes.auth import NoneAuth, get_auth
from archimedes.configuration import ArchimedesConstants  # config
from archimedes.data.api.forecast import (
    forecast_diff,
    forecast_get,
    forecast_get_by_ref_time_interval,
    forecast_list_ref_times,
)
from archimedes.data.api.intraday import get_intraday_trades
from archimedes.data.api.metadata import (
    list_ids,
    list_prediction_ids,
    list_series_price_areas,
)
from archimedes.data.api.observation import get, get_latest
from archimedes.data.api.prediction import (
    get_predictions,
    get_predictions_ref_dts,
    store_prediction,
    store_predictions,
)
from archimedes.data.api.rk_within_day import (
    rk_comparison_by_price,
    rk_comparison_by_probability,
    rk_within_day_directions,
    rk_within_day_distributions,
    rk_within_day_large_down_fps,
    rk_within_day_large_up_fps,
)
from archimedes.utils import compact_print, environ, full_print

__all__ = [
    "ArchimedesConstants",
    "get_auth",
    "NoneAuth",
    "compact_print",
    "full_print",
    "environ",
    "get",
    "get_intraday_trades",
    "get_latest",
    "list_ids",
    "list_series_price_areas",
    "get_predictions",
    "get_predictions_ref_dts",
    "list_prediction_ids",
    "store_prediction",
    "store_predictions",
    "forecast_list_ref_times",
    "forecast_get",
    "forecast_diff",
    "forecast_get_by_ref_time_interval",
    "rk_within_day_directions",
    "rk_comparison_by_probability",
    "rk_within_day_large_up_fps",
    "rk_within_day_large_down_fps",
    "rk_within_day_distributions",
    "rk_comparison_by_price",
]



"""
Query splitting strategy for best performance
"""
import re
from typing import Dict, List, Union

import numpy as np
import pandas as pd
from iteration_utilities import unique_everseen

from archimedes.utils.utils import get_end_date, get_start_date

from .config import (
    DATE_RANGE,
    DEFAULT_NUMBER_OF_SUB_IDS,
    NUMBER_OF_DATA_POINTS_PER_HOUR,
    PERMUTATION_RESULTS,
    SPLIT_REQUEST_AT_SIZE_BYTES,
    UNIT_DATA_SIZE_BY_SERIES,
)


def get_queries(  # pylint:disable=too-many-branches,too-many-locals
    series_ids: List[str],
    price_areas: Union[List[str], None],
    start: Union[str, pd.Timestamp, None],
    end: Union[str, pd.Timestamp, None],
    is_json_query: bool,
) -> List[Dict]:
    """
    Get split queries
    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (Union[List[str], None]): The price areas to pick.
        start (Union[str, pd.Timestamp, None])): The first datetime to request.
        end (Union[str, pd.Timestamp, NOne])): The last datetime to request.
        is_json_query (bool): True if the query is to be made for observations_json
                              or if it is to be made for data/get.

    Returns:
        List of split up queries.
    """
    if series_ids is None or len(series_ids) == 0:
        return []

    start, end = get_start_date(start), get_end_date(end)

    series_metadata = get_series_metadata(series_ids, price_areas, start, end)
    sum_approx_response_size_in_bytes = (
        -1 if series_metadata.empty else series_metadata["sum"]["approx_size_in_bytes"]
    )

    if sum_approx_response_size_in_bytes < SPLIT_REQUEST_AT_SIZE_BYTES:
        if not is_json_query:
            return [
                {
                    "start": start,
                    "end": end,
                    "series_ids": sorted(series_ids),
                    "price_areas": price_areas,
                }
            ]

        queries = [
            {
                "start": start,
                "end": end,
                "series_id": series_id,
                "price_areas": price_areas,
            }
            for series_id in series_ids
        ]
        return list(unique_everseen(queries))

    large_series_ids = []
    if not series_metadata.empty:
        for series_id in series_ids:
            if (
                series_metadata[series_id]["approx_size_in_bytes"]
                >= SPLIT_REQUEST_AT_SIZE_BYTES
            ):
                large_series_ids.append(series_id)

    large_queries = []
    for large_series_id in large_series_ids:
        duration = (end - start).total_seconds() / 2
        for offset in range(2):
            actual_start = start + pd.Timedelta(seconds=offset * duration)
            possible_end = actual_start + pd.Timedelta(seconds=duration)
            actual_end = min(end, possible_end)
            large_queries.extend(
                get_queries(
                    [large_series_id],
                    price_areas,
                    actual_start,
                    actual_end,
                    is_json_query,
                )
            )

    remaining_series_ids = list(set(series_ids) - set(large_series_ids))

    if is_json_query:
        remaining_queries = [
            {
                "start": start,
                "end": end,
                "series_id": series_id,
                "price_areas": price_areas,
            }
            for series_id in remaining_series_ids
        ]
    else:
        if len(remaining_series_ids) == 0:
            remaining_queries = []
        elif len(remaining_series_ids) == 1:
            remaining_queries = [
                {
                    "start": start,
                    "end": end,
                    "series_ids": sorted(remaining_series_ids),
                    "price_areas": price_areas,
                }
            ]
        else:
            if large_queries:
                # In this case, some large queries were contributing to the split
                # requirement but maybe the remaining queries don't need to be split
                # at all. Also, since remaining_queries is a subset of series_ids
                # without the large queries, it should never lead to infinite recursion.
                remaining_queries = get_queries(
                    remaining_series_ids, price_areas, start, end, is_json_query
                )
            else:
                remaining_queries = []
                for half_series_ids in np.array_split(np.sort(remaining_series_ids), 2):
                    remaining_queries.extend(
                        get_queries(
                            sorted(list(half_series_ids)),
                            price_areas,
                            start,
                            end,
                            is_json_query,
                        )
                    )

    return list(unique_everseen(large_queries + remaining_queries))


def get_queries_observation(
    series_ids: List[str],
    price_areas: List[str],
    start: Union[str, pd.Timestamp],
    end: Union[str, pd.Timestamp],
) -> List[Dict]:
    """
    Get split queries
    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (Union[List[str], None]): The price areas to pick.
        start (Union[str, pd.Timestamp, None])): The first datetime to request.
        end (Union[str, pd.Timestamp, NOne])): The last datetime to request.

    Returns:
        List[Dict]: List of split up queries.
    """
    return get_queries(series_ids, price_areas, start, end, False)


def get_queries_observation_json(
    series_ids: List[str],
    price_areas: List[str],
    start: Union[str, pd.Timestamp],
    end: Union[str, pd.Timestamp],
) -> List[Dict]:
    """
    Get split queries
    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (Union[List[str], None]): The price areas to pick.
        start (Union[str, pd.Timestamp, None])): The first datetime to request.
        end (Union[str, pd.Timestamp, NOne])): The last datetime to request.

    Returns:
        List[Dict]: List of split up queries.
    """
    return get_queries(series_ids, price_areas, start, end, True)


def get_series_metadata(
    series_ids: List[str],
    price_areas: List[str],
    start: Union[str, pd.Timestamp],
    end: Union[str, pd.Timestamp],
):
    """
    Get series metadata
    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (Union[List[str], None]): The price areas to pick.
        start (Union[str, pd.Timestamp, None])): The first datetime to request.
        end (Union[str, pd.Timestamp, NOne])): The last datetime to request.

    Returns:
        pd.DataFrame: A dataframe with information about the requested
                      range for the series.
    """
    series_metadata = {}

    for series_id in series_ids:
        actual_start, actual_end, hours_of_data = _get_hours_of_data(
            series_id, start, end
        )
        number_of_rows = _get_number_of_rows(series_id, hours_of_data)
        number_of_columns = _get_number_of_columns(series_id, price_areas)
        approx_size_in_bytes = _get_approx_size_in_bytes(
            series_id,
            number_of_columns,
            number_of_rows,
        )
        series_metadata[series_id] = {
            "actual_start": actual_start,
            "actual_end": actual_end,
            "hours_of_data": hours_of_data,
            "number_of_columns": number_of_columns,
            "number_of_rows": number_of_rows,
            "approx_size_in_bytes": approx_size_in_bytes,
        }

    df = pd.DataFrame.from_dict(series_metadata)
    df["sum"] = df.loc["approx_size_in_bytes"].sum()
    return df


def _get_hours_of_data(series_id, start, end):
    for match_string, (series_start_date, series_end_date) in DATE_RANGE.items():
        if re.match(match_string, series_id):
            hours_start = max(start, series_start_date)
            hours_end = min(end, series_end_date)
            hours_of_data = max((hours_end - hours_start).total_seconds() / 3600, 0)
            return hours_start, hours_end, hours_of_data
    return None, None, None


def _get_number_of_columns(series_id, price_areas):
    """
    This will return a wrong value if some price_areas do not exist for a series
    which needs_permutation.
    """
    r = _get_num_columns_needs_permutation(series_id)  # pylint:disable=invalid-name

    if price_areas and r is False:
        return len(price_areas)

    for match_string, num_columns in DEFAULT_NUMBER_OF_SUB_IDS.items():
        if re.match(match_string, series_id):
            if r is False:
                return num_columns
            perms = np.math.factorial(num_columns) / np.math.factorial(num_columns - r)
            if not price_areas:
                return perms
            perms_of_rest = np.math.factorial(
                num_columns - len(price_areas)
            ) / np.math.factorial(num_columns - len(price_areas) - r)
            net_perms = perms - perms_of_rest
            if net_perms >= 0:
                return net_perms
            return perms

    return None


def _get_num_columns_needs_permutation(series_id):
    for match_string, num_at_a_time in PERMUTATION_RESULTS.items():
        if re.match(match_string, series_id):
            return num_at_a_time

    return False


def _get_number_of_rows(series_id, hours_of_data):
    for match_string, num_rows_per_hour in NUMBER_OF_DATA_POINTS_PER_HOUR.items():
        if re.match(match_string, series_id):
            return num_rows_per_hour * hours_of_data

    return None


def _get_approx_size_in_bytes(series_id, number_of_columns, number_of_rows):
    for match_string, unit_size_in_bytes in UNIT_DATA_SIZE_BY_SERIES.items():
        if re.match(match_string, series_id):
            return number_of_columns * number_of_rows * unit_size_in_bytes

    return None



from typing import List

import pandas as pd

from archimedes.data.common import get_api_base_url_v2
from archimedes.utils import get_end_date, get_start_date
from archimedes.utils.api_request import make_api_request
from archimedes.utils.split import get_queries_observation_json
from archimedes.utils.threaded_executor import execute_many


def get_intraday_trades(
    price_areas: List[str] = None,
    start: str = None,
    end: str = None,
    *,
    access_token: str = None,
):
    """Get raw intraday trades from Archimedes Database

    This function can be used to fetch raw time series from the Archimedes Database
    without any post-processing.
    To see which series are available, use `list_ids()`.

    Example:
        >>> import archimedes
        >>> archimedes.get(
        >>>     price_areas=["NO1",],
        >>>     start="2020-06-20T04:00:00+00:00",
        >>>     end="2020-06-20T09:00:00+00:00",
        >>> )
                from_dt to_dt   series_id   version price   trade_time  buy_area    sell_area   attributes
        0	2022-03-01T00:00:00+00:00	2022-03-01T01:00:00+00:00	NP/IntradayTrades	1646193645	146.63	2022-02-28T17:02:06.934+0000	OPX	NO1	{'price': 146.63,...,  'product_code': 'PH-20220301-01'}
        1	2022-03-01T00:00:00+00:00	2022-03-01T01:00:00+00:00	NP/IntradayTrades	1646193645	146.63	2022-02-28T17:02:06.934+0000	OPX	NO1	{'price': 146.63,... , 'product_code': 'PH-20220301-01'}
        2	2022-03-01T00:00:00+00:00	2022-03-01T01:00:00+00:00	NP/IntradayTrades	1646193645	146.63	2022-02-28T17:02:06.934+0000	OPX	NO1	{'price': 146.63, ..., 'product_code': 'PH-20220301-01'}
        ...
        155	2022-03-01T23:00:00+00:00	2022-03-02T00:00:00+00:00	NP/IntradayTrades	1646193645	148.99	2022-03-01T13:05:40.934+0000	NL	NO1	{'price': 148.99,... , 'product_code': 'PH-20220301-24'}
        156	2022-03-01T23:00:00+00:00	2022-03-02T00:00:00+00:00	NP/IntradayTrades	1646193645	148.8	2022-03-01T12:57:58.777+0000	NO1	FI	{'price': 148.8,...,  'product_code': 'PH-20220301-24'}
        157	2022-03-01T23:00:00+00:00	2022-03-02T00:00:00+00:00	NP/IntradayTrades	1646193645	148.8	2022-03-01T12:58:02.115+0000	NO1	FI	{'price': 148.8, ... , 'product_code': 'PH-20220301-24'}


    Args:
        price_areas (List[str], optional): The price areas to pick, all price areas if None. Defaults to None.
        start (str, optional): The first datetime to fetch (inclusive). Returns all if None. Defaults to None.
        end (str, optional): The last datetime to fetch (exclusive). Returns all if None. Defaults to None.
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with all the time series data

    Raises:
        HTTPError: If an HTTP error occurs when requesting the API.
        NoneAuth: If the user is unauthorized or if the authorization has expired.
    """  # pylint:disable=line-too-long

    if isinstance(price_areas, str):
        price_areas = [price_areas]

    start, end = get_start_date(start), get_end_date(end)

    queries = get_queries_observation_json(
        ["NP/IntradayTrades"], price_areas, start, end
    )

    base_url = get_api_base_url_v2()

    params_array = [
        dict(
            url=f"{base_url}/observation_json/get",
            access_token=access_token,
            params=query,
        )
        for query in queries
    ]

    observation_data = execute_many(make_api_request, params_array)

    if len(observation_data) == 0:
        return pd.DataFrame(
            columns=[
                "from_dt",
                "to_dt",
                "series_id",
                "version",
                "price",
                "trade_time",
                "buy_area",
                "sell_area",
                "attributes",
            ]
        )
    observation_data = [
        {
            **i,
            "price": i["value"].get("price"),
            "trade_time": i["value"].get("trade_time"),
        }
        for i in observation_data
    ]
    observation_data = pd.DataFrame.from_dict(observation_data)

    # Extracting price area and series id
    observation_data[
        ["series_id1", "series_id2", "buy_area", "sell_area"]
    ] = observation_data["series_id"].str.split("/", 3, expand=True)
    observation_data["series_id"] = (
        observation_data["series_id1"] + "/" + observation_data["series_id2"]
    )
    observation_data["attributes"] = observation_data["value"]
    observation_data["attributes_str"] = observation_data["value"].astype(str)
    observation_data = observation_data.drop_duplicates(
        ["from_dt", "series_id", "attributes_str"]
    )
    observation_data = observation_data.drop(
        ["series_id1", "series_id2", "value", "version", "attributes_str"], axis=1
    )
    observation_data = observation_data.sort_values(by=["from_dt"]).reset_index(
        drop=True
    )

    return observation_data



from .environ_context import environ
from .format import compact_print, full_print
from .utils import (
    composite_id_to_attributes,
    create_composite_id,
    get_end_date,
    get_start_date,
)



from msal_extensions import (
    FilePersistence,
    PersistedTokenCache,
    build_encrypted_persistence,
)

from .configuration import get_msal_path


def build_persistence(location, fallback_to_plaintext=False):
    """Build a suitable persistence instance based your current OS"""
    try:
        return build_encrypted_persistence(location)
    except:
        if not fallback_to_plaintext:
            raise
        return FilePersistence(location)


def get_token_cache():
    persistence = build_persistence(get_msal_path(), True)
    return PersistedTokenCache(persistence)



from typing import Dict

import pandas as pd

from archimedes import ArchimedesConstants


def create_composite_id(
    series_id: str,
    price_area: str = None,
    quantile: str = None,
    hours_ahead: pd.Timedelta = None,
) -> str:
    """Create a composite ID for a time series.

    The time series can be a data series, or a prediction series.
    For data series, eg. Nordpool data, the composite ID will be similar to:

        NP/ConsumptionImbalancePrices:NO2

    where NP/ConsConsumptionImbalancePrices is the series_id, and NO2 is the price_area.

    For predictions, the composite ID will be something like:
        PX/rk-naive:NO2:H4:Q50

    where PX/rk-naive is the series_id, NO2 is the price area,
    H4 is "four hours ahead" and Q50 is quantile 50.

    Example:
        >>> import pandas as pd
        >>> h = pd.Timedelta(3, "hours")
        >>> create_composite_id(
        >>>     series_id="PX/my-prediction",
        >>>     price_area="NO2",
        >>>     hours_ahead=h,
        >>>     quantile=50
        >>> )
        ""
        {
            "series_id": "PX/my-prediction",
            "price_area": "NO2",
            "hours_ahead": 3,
            "quantile": 50
        }

    Args:
        series_id (str): The series ID.
        price_area (str, optional): The price area. Defaults to None.
        quantile (str, optional): The quantile. Defaults to None.
        hours_ahead (pd.Timedelta, optional):
            Number of hours the timestamp of the predicted value is ahead of the
            reference timestamp for the prediction.

    Returns:
        str: The composite id
    """
    if isinstance(hours_ahead, pd.Timedelta):
        # Converting the timedelta to an integer
        hours_ahead_ = int(hours_ahead.seconds / 3600)
    elif hours_ahead is None:
        pass
    else:
        raise TypeError(
            f"Invalid type {type(hours_ahead)} for hours_ahead: {hours_ahead}>.\n"
            + "Consider using pd.to_timedelta() to convert."
        )

    ordered_parts = [
        series_id,
        price_area,
        f"H{hours_ahead_}" if hours_ahead else None,
        f"Q{quantile}" if quantile else None,
    ]

    nones_removed = [part for part in ordered_parts if part is not None]

    composite_id = ":".join(nones_removed)

    return composite_id


def composite_id_to_attributes(composite_id: str) -> Dict:
    """Expand a composite ID to a dict of attributes.

    Since we will often be working with the composite IDs for the
    time series – and since they are an easy way to carry metadata
    about the predictions around, it is useful to be able to do the
    reverse operation of `create_composite_id`.

    A dataframe with columns:
        (dt, composite_id, value)
    can then easily be converted into a dataframe with columns:
        (dt, series_id, hours_ahead, quantile, ...)

    Example (simple):
        >>> composite_id = "PX/rk-nn:NO2:H3:Q90"
        >>> composite_id_to_attribute(composite_id)
        {"series_id": "PX/rk-nn", "price_area": "NO2", "hours_ahead": 3, "quantile": 90}

    Example (dataframe):
        >>> df = pd.DataFrame(
        >>>     [
        >>>         ["17:00", "PX/rk-nn:NO2:H3:Q50", 28.43],
        >>>         ["17:00", "PX/rk-nn:NO2:H3:Q90", 31.11]
        >>>     ],
        >>>     columns=["from_dt", "composite_id", "value"]
        >>> )
        >>> dfa = pd.concat(
        >>>     [
        >>>         df,
        >>>         df["composite_id"].apply(
        >>>             composite_id_to_attributes
        >>>         ).apply(
        >>>             pd.Series
        >>>         )
        >>>     ],
        >>>     axis=1
        >>> )
        ...

    Args:
        composite_id (str): The composite ID

    Returns:
        Dict[]: A dictionary of all the attributes
    """
    attributes = {}
    elements = composite_id.split(":")
    series_id, *rest_of_the_series_id = elements
    attributes["series_id"] = series_id

    for series_id_part in rest_of_the_series_id:
        # Hours ahead part ->
        if "H" in series_id_part:
            attributes["hours_ahead"] = pd.Timedelta(int(series_id_part[1:]), "hours")
        # Quantile part ->
        elif "Q" in series_id_part:
            attributes["quantile"] = int(series_id_part[1:])
        # Price area parts ->
        elif "NO" in series_id_part:
            attributes["price_area"] = series_id_part
        elif "SE" in series_id_part:
            attributes["price_area"] = series_id_part
        elif "FI" in series_id_part:
            attributes["price_area"] = series_id_part
        elif "DK" in series_id_part:
            attributes["price_area"] = series_id_part
        else:
            raise ValueError(f"Uncaught attribute: {series_id_part}")

    return attributes


def get_start_date(start):
    """
    Args:
        start (str): The start date provided by the user

    Returns:
        pd.Timestamp: start date given in string form to pd.Timestamp. If not given,
                      or empty, it returns ArchimedesConstants.DATE_LOW
    """
    return pd.to_datetime(start, utc=True) if start else ArchimedesConstants.DATE_LOW


def get_end_date(end):
    """
    Args:
        end (str): The end date provided by the user

    Returns:
        pd.Timestamp: end date given in string form to pd.Timestamp. If not given,
                      or empty, it returns ArchimedesConstants.DATE_HIGH
    """
    return pd.to_datetime(end, utc=True) if end else ArchimedesConstants.DATE_HIGH



import os
from contextlib import contextmanager


@contextmanager
def environ(env):
    """Temporarily set environment variables inside the context manager and
    fully restore previous environment afterwards

    Borrowed from here: https://gist.github.com/igniteflow/7267431

    Example usage:
    print(os.environ['USER']) # prints username of current user

    user_temporary = "temp_user"
    with environ({"USER": user_temporary}):
        print(os.environ['USER']) # prints "temp_user"

    print(os.environ['USER']) # prints username of current user
    """
    original_env = {key: os.getenv(key) for key in env}
    os.environ.update(env)
    try:
        yield
    finally:
        for key, value in original_env.items():
            if value is None:
                del os.environ[key]
            else:
                os.environ[key] = value



import abc
import os

import pandas as pd

# Global archimedes config
USER_HOME_DIR = os.path.expanduser("~")
ARCHIMEDES_CONF_DIR = os.path.join(USER_HOME_DIR, ".archimedes")


def get_environment():
    env = os.environ.get("ARCHIMEDES_ENVIRONMENT")
    if env is None:
        env = os.environ.get("ENVIRONMENT", "prod")
    return env.strip().lower()


def get_msal_path():
    msal_path = os.path.join(ARCHIMEDES_CONF_DIR, f"msal-{get_environment()}.cache.bin")
    return msal_path


ARCHIMEDES_API_CONFIG = {
    "prod": {
        "client_id": "5bc3a702-d753-43ff-9051-e7fdfdd95023",
        "aad_app_client_id": "c0a9f773-6276-4d71-8df6-7239e695aff6",
        "url": "https://api.fabapps.io",
        "authority": "https://login.microsoftonline.com/common",
    },
    "dev": {
        "client_id": "2e2f3c84-d1aa-49cc-90a2-fd3fa3380d27",
        "aad_app_client_id": "eaaa9f9f-395d-46aa-847f-b5fb6c087ff6",
        "url": "https://api-dev.fabapps.io",
        "authority": "https://login.microsoftonline.com/common",
    },
}


class ArchimedesConstants:
    DATE_LOW = pd.to_datetime("1900-01-01T00:00:00+00:00")
    DATE_HIGH = pd.to_datetime("2090-01-01T00:00:00+00:00")


class InvalidEnvironmentException(Exception):
    pass


class ApiConfig(abc.ABC):
    def __init__(self, env):
        self.config = ARCHIMEDES_API_CONFIG
        if env not in self.config.keys():
            raise InvalidEnvironmentException(
                f"Invalid environment {env}, "
                f"supported values are {', '.join([str(key) for key in self.config.keys()])}"
            )
        self.environment = env.lower()

    def __getattr__(self, item):
        env_config = self.config[self.environment]
        return env_config[item]


def get_api_config():
    return ApiConfig(get_environment())


def get_api_base_url(api_version: int) -> str:
    return f"{get_api_config().url}/v{api_version}"



import pandas as pd

SPLIT_REQUEST_AT_SIZE_BYTES = 104857600  # 100 MB

# The following values are calculated by benchmarking request payload sizes
UNIT_DATA_SIZE_BY_SERIES = {
    r"NP\/BuyCurve": 23630,
    r"NP\/SellCurve": 23630,
    r"NP\/IntradayTrades": 520,
    r".*": 170,
}

# For the following, the first match is the one used, so it needs to be ordered
# correctly
# The values are roughly equal to len(archimedes.list_series_price_areas(series_id))
DEFAULT_NUMBER_OF_SUB_IDS = {
    r"IV\/.*": 1,
    r"MET\/forecast_wind_speed_10m_at_windfarms": 11,
    r"MET\/observation_wind_speed_10m_at_windfarms": 11,
    r"MET\/.*": 18,
    r"NP\/SysPrices": 1,
    r"NP\/AreaBlockPricesPeak": 15,
    r"NP\/AutomaticReserveDownVolume": 11,
    r"NP\/AutomaticReserveUpVolume": 11,
    r"NP\/BuyCurve": 1,
    r"NP\/Capacities": 10,
    r"NP\/CapacitiesPrognosis": 28,
    r"NP\/ConsumptionPrognosis": 18,
    r"NP\/FlowsAtAreaPrice": 28,
    r"NP\/FlowsAtSystemPrice": 5,
    r"NP\/IntradayTrades": 30,
    r"NP\/NetExchange": 7,
    r"NP\/PhysicalCapacities": 6,
    r"NP\/ProductionTotals": 18,
    r"NP\/ProductionTotalPrognosis": 17,
    r"NP\/ProductionWindPower": 17,
    r"NP\/ProductionWindPrognosis": 16,
    r"NP\/ReductionReasons": 16,
    r"NP\/SellCurve": 1,
    r"NP\/SystemPrices": 1,
    r"NP\/TotalConsumption": 18,
    r"NP\/TurnoverAtSystemPrice": 1,
    r"NP\/VolumeAcceptedBlocksBuy": 1,
    r"NP\/VolumeAcceptedBlocksSell": 1,
    r"NP\/.*": 12,
    r"OMX\/.*": 1,
    r"SN\/ConsumptionTotal": 8,
    r"SN\/FCRWVolume": 1,
    r"SN\/FlowExport": 1,
    r"SN\/FlowImport": 1,
    r"SN\/FlowNet": 1,
    r"SN\/ProductionHydro": 6,
    r"SN\/ProductionNuclear": 3,
    r"SN\/ProductionThermal": 8,
    r"SN\/ProductionTotal": 8,
    r"SN\/ProductionUnspecifiedSource": 5,
    r"SN\/ProductionWind": 8,
    r"SN\/WeeklyMTS.*Time": 1206,
    r"SN\/.*": 5,
    r".*": 12,
}

# The series_ids that appear as a result fo the following SQL queries:
# with s2 as (
# 	with s1 as (
# 		select
# 	    	subpath(series_id, 0, 2) as series_prefix,
# 	       	ltree2text(series_id) as sid
# 	    from observations --this should also be done for observations_json and forecasts
# 	    order by from_dt desc
# 	)
# 	    select
# 	    	series_prefix,
# 	    	sid,
# 	    	array_length(string_to_array(sid, '.'), 1) - 1 as c
# 	    from s1
# )
# select distinct(series_prefix) from s2 where c>2
PERMUTATION_RESULTS = {
    r"NP\/Capacities": 2,
    r"NP\/CapacitiesPrognosis": 2,
    r"NP\/FlowsAtAreaPrice": 2,
    r"NP\/FlowsAtSystemPrice": 2,
    r"NP\/IntradayTrades": 2,
    r"NP\/PhysicalCapacities": 2,
    r"NP\/ReductionReasons": 2,
}

# The following values come from estimating how many data points per series_ids exist
NUMBER_OF_DATA_POINTS_PER_HOUR = {
    r"NP\/CapacitiesPrognosis": 0.046,
    r"NP\/IntradayTrades": 4,
    r"OMX\/.*": 0.00034,
    r"SN\/WeeklyMTS.*Time": 1 / (7 * 24),
    r".*": 1,
}

END_OF_TODAY = pd.Timestamp.utcnow().replace(
    hour=0, minute=0, second=0, microsecond=0
) + pd.Timedelta(days=1)

DATE_RANGE = {
    r"IV\/BrentOil": (
        pd.to_datetime("2000-01-04 22:00:00+00"),
        END_OF_TODAY,
    ),
    r"IV\/CarbonEmissionsEUA": (
        pd.to_datetime("2008-07-04 16:00:00+00"),
        END_OF_TODAY,
    ),
    r"IV\/Coal": (
        pd.to_datetime("2008-12-08 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"IV\/NaturalGas": (
        pd.to_datetime("2000-01-04 22:00:00+00"),
        END_OF_TODAY,
    ),
    r"MET\/forecast_.*": (
        pd.to_datetime("2019-11-26 13:00:00+00"),
        END_OF_TODAY,
    ),
    r"MET\/.*": (
        pd.to_datetime("2013-09-01 00:00:00+00"),
        END_OF_TODAY,
    ),
    r"NP\/BuyCurve": (
        pd.to_datetime("2013-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"NP\/FlowsAtSystemPrice": (
        pd.to_datetime("2015-12-31 23:00:00+00"),
        pd.to_datetime("2019-01-08 22:00:00+00"),
    ),
    r"NP\/HydroReservoir.*": (
        pd.to_datetime("2017-01-01 00:00:00+00"),
        END_OF_TODAY,
    ),
    r"NP\/IntradayTrades": (
        pd.to_datetime("2010-01-01 00:00:00+00"),
        END_OF_TODAY,
    ),
    r"NP\/NetFlowsAtSystemPrice": (
        pd.to_datetime("2013-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"NP\/SellCurve": (
        pd.to_datetime("2013-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"NP\/VolumeAcceptedBlocksBuy": (
        pd.to_datetime("2013-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"NP\/VolumeAcceptedBlocksSell": (
        pd.to_datetime("2013-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"NP\/.*": (
        pd.to_datetime("2015-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/Consumption": (
        pd.to_datetime("2000-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/ConsumptionTotal": (
        pd.to_datetime("2022-04-25 12:46:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/FCRNPrice": (
        pd.to_datetime("2013-01-07 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/FCRNVolume": (
        pd.to_datetime("2013-01-07 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/FCRWVolume": (
        pd.to_datetime("2013-01-11 23:00:00+00"),
        pd.to_datetime("2021-01-10 22:00:00+00"),
    ),
    r"SN\/FlowExport": (
        pd.to_datetime("2018-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/FlowImport": (
        pd.to_datetime("2018-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/FlowNet": (
        pd.to_datetime("2018-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/ProductionHydro": (
        pd.to_datetime("2022-04-25 12:46:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/ProductionNuclear": (
        pd.to_datetime("2022-04-25 12:46:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/ProductionThermal": (
        pd.to_datetime("2022-04-25 12:46:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/ProductionTotal": (
        pd.to_datetime("2022-04-25 12:46:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/ProductionUnspecifiedSource": (
        pd.to_datetime("2022-04-25 12:46:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/ProductionWind": (
        pd.to_datetime("2022-04-25 12:46:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/Production": (
        pd.to_datetime("2000-12-31 23:00:00+00"),
        END_OF_TODAY,
    ),
    r"SN\/WeeklyMTS.*Time": (
        pd.to_datetime("2007-01-01 00:00:00+00"),
        pd.Timestamp.utcnow() + pd.Timedelta(days=7),
    ),
    r"OMX\/.*": (
        pd.to_datetime("2022-02-28 23:00:00+00"),
        pd.Timestamp.utcnow() + pd.Timedelta(days=10 * 365),
    ),
    r".*": (
        pd.to_datetime("2010-01-01 00:00:00+00"),
        END_OF_TODAY,
    ),
}



import pandas as pd

from archimedes.data.common import get_api_base_url_v2
from archimedes.utils.api_request import make_api_request


def rk_within_day_directions(
    start: str,
    end: str,
    price_area: str,
    ref_dt: str = None,
    *,
    access_token: str = None,
):
    """
    Method to get directions for rk within day
    Parameters:
        start: Start timestamp
        end: end timestamp
        price_area: Price area
        ref_dt: Reference timestamp
        access_token: Access token
    """
    query = {
        "start": start,
        "end": end,
        "price_area": price_area,
    }
    if ref_dt is not None:
        query["ref_dt"] = ref_dt
    data = make_api_request(
        f"{get_api_base_url_v2()}/rk_within_day/directions",
        access_token=access_token,
        params=query,
    )
    return pd.DataFrame.from_dict(data)


def rk_within_day_distributions(
    start: str,
    end: str,
    price_area: str,
    ref_dt: str = None,
    *,
    access_token: str = None,
):
    """
    Method to get distributions for rk within day
    Parameters:
        start: Start timestamp
        end: end timestamp
        price_area: Price area
        ref_dt: Reference timestamp
        access_token: Access token
    """
    query = {
        "start": start,
        "end": end,
        "price_area": price_area,
    }
    if ref_dt is not None:
        query["ref_dt"] = ref_dt
    data = make_api_request(
        f"{get_api_base_url_v2()}/rk_within_day/distributions",
        access_token=access_token,
        params=query,
    )
    return pd.DataFrame.from_dict(data)


def rk_within_day_large_up_fps(
    start: str,
    end: str,
    price_area: str,
    ref_dt: str = None,
    conditional: bool = False,
    *,
    access_token: str = None,
):
    """
    Method to get large up for rk within day
    Parameters:
        start: Start timestamp
        end: end timestamp
        price_area: Price area
        ref_dt: Reference timestamp
        conditional: If set to False, ignores the direction of RK
        access_token: Access token
    """
    query = {
        "start": start,
        "end": end,
        "price_area": price_area,
        "conditional": conditional,
    }
    if ref_dt is not None:
        query["ref_dt"] = ref_dt
    data = make_api_request(
        f"{get_api_base_url_v2()}/rk_within_day/large_up_fps",
        access_token=access_token,
        params=query,
    )
    return pd.DataFrame.from_dict(data)


def rk_within_day_large_down_fps(
    start: str,
    end: str,
    price_area: str,
    ref_dt: str = None,
    conditional: bool = False,
    *,
    access_token: str = None,
):
    """
    Method to get large down for rk within day
    Parameters:
        start: Start timestamp
        end: end timestamp
        price_area: Price area
        ref_dt: Reference timestamp
        conditional: If set to False, ignores the direction of RK
        access_token: Access token
    """
    query = {
        "start": start,
        "end": end,
        "price_area": price_area,
        "conditional": conditional,
    }
    if ref_dt is not None:
        query["ref_dt"] = ref_dt
    data = make_api_request(
        f"{get_api_base_url_v2()}/rk_within_day/large_down_fps",
        access_token=access_token,
        params=query,
    )
    return pd.DataFrame.from_dict(data)


def rk_comparison_by_price(
    start: str,
    price_area: str,
    ref_price: int,
    *,
    access_token: str = None,
):
    """
    Method to get probability for rk within day in comparison to provided price
    Parameters:
        start: Start timestamp
        price_area: Price area
        ref_price: Reference price to compare against
        access_token: Access token
    """
    query = {
        "start_time": start,
        "price_area": price_area,
        "ref_price": ref_price,
    }
    data = make_api_request(
        f"{get_api_base_url_v2()}/rk_within_day/price_comparison/by_price",
        access_token=access_token,
        params=query,
    )
    return pd.DataFrame.from_dict(data)


def rk_comparison_by_probability(
    start: str,
    price_area: str,
    ref_probability: int,
    *,
    access_token: str = None,
):
    """
    Method to get probability for rk within day in comparison to provided price
    Parameters:
        start: Start timestamp
        price_area: Price area
        ref_probability: Reference probability to compare against
        access_token: Access token
    """
    query = {
        "start_time": start,
        "price_area": price_area,
        "ref_probability": ref_probability,
    }
    data = make_api_request(
        f"{get_api_base_url_v2()}/rk_within_day/price_comparison/by_probability",
        access_token=access_token,
        params=query,
    )
    return pd.DataFrame.from_dict(data)



import os

import msal

from .configuration import get_api_config
from .token_cache import get_token_cache


def get_scopes():
    api_app_id_uri_base = f"api://{get_api_config().aad_app_client_id}"
    return [
        f"{api_app_id_uri_base}/.default",
    ]


class NoneAuth(Exception):
    """User not logged in. Please log in using `arcl auth login <organization_id>"""

    pass


class ArchimedesPublicClientAuth:
    def __init__(self, client_id, authority, cache=None):
        self.app = self.build_msal_app(client_id, authority, cache=cache)

    @staticmethod
    def build_msal_app(client_id, authority, cache=None):
        return msal.PublicClientApplication(
            client_id,
            authority=authority,
            token_cache=cache,
        )


class ArchimedesLocalAuth(ArchimedesPublicClientAuth):
    def __init__(self):
        api_config = get_api_config()
        super().__init__(api_config.client_id, api_config.authority, get_token_cache())

    def get_access_token_silent(self):
        # We now check the cache to see
        # whether we already have some accounts that the end user already used to sign in before.
        accounts = self.app.get_accounts()
        if not accounts:
            return None

        chosen = accounts[0]
        result = self.app.acquire_token_silent(get_scopes(), account=chosen)

        if result is None or "access_token" not in result:
            description = (
                f" Error details: {result['error_description']}"
                if result is not None and "error_description" in result
                else ""
            )
            raise NoneAuth(
                f"User not logged in. Please log in using `arcl auth login <organization_id>`.{description}"
            )

        return result.get("access_token")


class ArchimedesConfidentialAuth:
    def __init__(self, client_id, client_credential, authority):
        self.app = self.build_confidential_msal_app(
            client_id, client_credential, authority
        )

    def get_access_token_silent(self):
        result = self.app.acquire_token_for_client(get_scopes())

        if result is None or "access_token" not in result:
            description = (
                f" Error details: {result['error_description']}"
                if result is not None and "error_description" in result
                else ""
            )
            raise NoneAuth(
                "Authentication failed. Please make sure that AZURE_AD_APP_ID, AZURE_AD_APP_CLIENT_CREDENTIAL and "
                f"AZURE_AD_TENANT_ID are properly configured.{description}"
            )

        return result.get("access_token")

    @staticmethod
    def build_confidential_msal_app(client_id, client_credential, authority):
        return msal.ConfidentialClientApplication(
            client_id=client_id,
            client_credential=client_credential,
            authority=authority,
        )


def get_auth():
    archimedes_auth = None
    use_app_authentication = os.getenv(
        "USE_APP_AUTHENTICATION", "false"
    ).strip().lower() not in ["false", "f", "0"]
    use_web_authentication = os.getenv(
        "USE_WEB_AUTHENTICATION", "false"
    ).strip().lower() not in ["false", "f", "0"]
    assert not (
        use_app_authentication and use_web_authentication
    ), "Only one of USE_APP_AUTHENTICATION or USE_WEB_AUTHENTICATION can be set to TRUE"
    is_local = not use_app_authentication and not use_web_authentication
    if is_local:
        archimedes_auth = ArchimedesLocalAuth()
    else:
        azure_ad_tenant_id = os.getenv("AZURE_AD_TENANT_ID")
        azure_ad_client_id = os.getenv("AZURE_AD_APP_ID")
        azure_ad_authority = f"https://login.microsoftonline.com/{azure_ad_tenant_id}"

        if use_app_authentication:
            azure_ad_client_credential = os.getenv("AZURE_AD_APP_CLIENT_CREDENTIAL")
            archimedes_auth = ArchimedesConfidentialAuth(
                azure_ad_client_id, azure_ad_client_credential, azure_ad_authority
            )
        elif use_web_authentication:
            archimedes_auth = None
    return archimedes_auth



import sys
from concurrent.futures import ThreadPoolExecutor


def default_merge_function(result, aggr):
    """
    Default callback function for execute_many to merge the results

    Args:
        result: result of the execution
        aggr: aggregate to merge to

    Returns:
        Aggregated result
    """
    if aggr is None:
        aggr = []
    aggr.extend(result)
    return aggr


def execute_many(func, params_array, merge_function=default_merge_function):
    """
    Execute a function multiple times in parallel for params in params_array and merge
    the results using the merge_function.

    Args:
        func: function to execute
        params_array: array of parameters for the function
        merge_function: function to merge the result

    Returns:
        Aggregated result of multiple executions
    """
    aggr = None
    with ThreadPoolExecutor() as executor:
        try:
            futures = [
                executor.submit(
                    func,
                    **params,
                )
                for params in params_array
            ]

            for future in futures:
                aggr = merge_function(future.result(), aggr)
        except KeyboardInterrupt:
            print("Cancelling tasks")
            cancelled_count = 0
            total_count = 0
            running_count = 0
            for future in futures:
                total_count += 1
                if future.running():
                    running_count += 1
                if future.cancel():
                    cancelled_count += 1
            print(
                f"Cancelled {cancelled_count}/{total_count} tasks. "
                f"Waiting for {running_count} running tasks to complete..."
            )
            sys.exit()
    return aggr



"""
A collection of functions for integrating with the Archimedes API
"""
import json
import os
from http import HTTPStatus

import requests
from requests.exceptions import (  # pylint:disable=redefined-builtin
    ConnectionError,
    ConnectTimeout,
    HTTPError,
    JSONDecodeError,
    Timeout,
)
from retry import retry

from archimedes import NoneAuth, get_auth

DEFAULT_TIMEOUT = int(os.getenv("ARCHIMEDES_API_TIMEOUT", "120"))  # 2 minutes
RETRY_EXCEPTIONS = (ConnectionError, ConnectTimeout, HTTPError, Timeout)


@retry(RETRY_EXCEPTIONS, tries=3, delay=2)
def make_api_request(url, method="GET", access_token=None, **kwargs):
    """
    Make request to the Archimedes API. It automatically retries 3 times on failure.

    Args:
        url:
            full URL of the API endpoint.
        method:
            HTTP method to use.
        access_token:
            authorization token; if None, tries to get the token automatically based on
            the authentication configuration.
        **kwargs:
            other kwargs to requests.request.

    Returns:
        JSON response
    """
    if access_token is None:
        archimedes_auth = get_auth()
        if archimedes_auth is None:
            raise NoneAuth(
                "access_token parameter must be passed when using "
                "USE_WEB_AUTHENTICATION"
            )
        access_token = archimedes_auth.get_access_token_silent()

    timeout = kwargs.pop("timeout", None)
    if timeout is None:
        timeout = DEFAULT_TIMEOUT
    kwargs["timeout"] = timeout

    headers = kwargs.pop("headers", None)
    if headers is None:
        headers = {}

    headers.update({"Authorization": f"Bearer {access_token}"})
    response = requests.request(method, url, headers=headers, **kwargs)

    if response.status_code not in [HTTPStatus.OK, HTTPStatus.CREATED]:
        try:
            response_json = response.json()
            if "message" in response_json:
                error_message = response_json.get("message")
            elif "detail" in response_json:
                error_message = response_json.get("detail")
            else:
                error_message = json.dumps(response_json)
        except JSONDecodeError:
            error_message = response.content
        params_str = json.dumps(kwargs.get("params"))
        raise HTTPError(
            f"API Error while requesting {url} with parameters {params_str}: "
            f"{error_message}"
        )

    return response.json()



import pandas as pd

from archimedes.data.common import get_api_base_url_v2
from archimedes.utils.api_request import make_api_request


def list_series_price_areas(series_id: str, *, access_token: str = None):
    """
    Retrieve all the price_areas which are available for the specified data series

    Example:
        >>> import archimedes
        >>> archimedes.list_series_price_areas('NP/AreaPrices')
           price_areas
        0          DK1
        1          DK2
        ...        ...
        10         SE3
        11         SE4

    Returns:
        Dataframe with all available price areas for the specified series_id
    """
    query = {
        "series_id": series_id,
    }
    base_url = get_api_base_url_v2()
    data = make_api_request(
        f"{base_url}/data/list_series_price_areas",
        access_token=access_token,
        params=query,
    )
    data = pd.DataFrame.from_dict(data)

    observation_data = make_api_request(
        f"{base_url}/observation_json/list_series_price_areas",
        access_token=access_token,
        params=query,
    )
    observation_data = pd.DataFrame.from_dict(observation_data)

    price_area_df = pd.concat([data, observation_data]).drop_duplicates()
    price_area_df = price_area_df.sort_values("price_areas").reset_index(drop=True)
    return price_area_df


def list_ids(sort: bool = False, *, access_token: str = None):
    """List all the series ids available.

    Example:
        >>> import archimedes
        >>> archimedes.list_ids()
                                    series_id
        0   NP/NegativeProductionImbalancePrices
        1                    NP/ProductionTotals
        ..                                   ...
        38                 NP/OrdinaryDownVolume
        39                    NP/SpecialUpVolume

    Args:
        sort (bool): False - return all series in one dataframe column, True - order
                             dataframe by data-origin
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with all available list_ids
    """
    base_url = get_api_base_url_v2()
    data = make_api_request(f"{base_url}/data/list_ids", access_token=access_token)
    data = pd.DataFrame.from_dict(data)

    observation_data = make_api_request(
        f"{base_url}/observation_json/list_ids",
        access_token=access_token,
    )
    observation_data = pd.DataFrame.from_dict(observation_data)

    series_df = pd.concat([data, observation_data]).drop_duplicates()
    series_df = series_df.sort_values(["series_id"]).reset_index(drop=True)
    if not sort:
        return series_df

    series_df["pre"] = series_df["series_id"].str.split("/", 1).str[0]
    series_df = pd.DataFrame.from_dict(
        series_df.groupby("pre")["series_id"].apply(list).to_dict(), orient="index"
    ).transpose()
    series_df = series_df[sorted(series_df.columns)]

    series_df = series_df.fillna("")
    return series_df.copy()


def list_prediction_ids(*, access_token: str = None):
    """List all the prediction series ids available.

    Example:
        >>> import archimedes
        >>> archimedes.list_prediction_ids()
                                     series_id
        0               PX/rk-nn-probabilities
        1   PX/rk-nn-direction-probabilities/U
        ..                                ...
        22                           PX/rk-901
        23                         PX/rk-naive
    """

    data = make_api_request(
        f"{get_api_base_url_v2()}/data/list_prediction_ids",
        access_token=access_token,
    )

    return pd.DataFrame.from_dict(data)



"""
A collection of functions for integrating with the Archimedes API
"""
from typing import List

import pandas as pd

from archimedes.data.common import get_api_base_url_v2
from archimedes.utils import get_end_date, get_start_date
from archimedes.utils.api_request import make_api_request
from archimedes.utils.threaded_executor import execute_many


def forecast_list_ref_times(
    series_id: str,
    start: pd.Timestamp = None,
    end: pd.Timestamp = None,
    limit: int = None,
    *,
    access_token: str = None,
):
    """
    List all forecast reference times (the times that the forecast was generated)

    Args:
        series_id (str):
            The ID of the data series to find all the forecast reference times (the
            time the forecast was generated). Retrieve the complete list of series
            (both forecasts and observations) using the list_ids resource.
        start (pd.Timestamp, optional):
            The first datetime to fetch (inclusive). Returns all if not set. Should be
            specified in ISO 8601 format.
            (eg - '2021-11-29T06:00:00+00:00')
        end (pd.Timestamp, optional):
            The last datetime to fetch (exclusive). Returns all if not set. Should be
            specified in ISO 8601 format.
            (eg - '2021-11-30T06:00:00+00:00')
        limit (int, optional):
            Limit the output to a specific number of entries. No limit if not specified.
        access_token (str, optional): Access token for the API

    Example:
        >>> import archimedes
        >>> start = pd.Timestamp("2022-01-09T06:00:00+00:00")
        >>> end = pd.Timestamp("2022-01-10T06:00:00+00:00")
        >>> archimedes.forecast_list_ref_times('MET/forecast_air_temperature_2m', start, end)
                                   ref_times
        0   2022-01-10T05:00:00.000000+00:00
        1   2022-01-10T04:00:00.000000+00:00
        ...                              ...
        22  2022-01-09T07:00:00.000000+00:00
        23  2022-01-09T06:00:00.000000+00:00

    Returns:
        Dataframe with all of the forecast reference times
    """  # pylint:disable=line-too-long
    query = {
        "series_id": series_id,
        "start": start.isoformat() if start is not None else None,
        "end": end.isoformat() if end is not None else None,
        "limit": limit,
    }

    data = make_api_request(
        f"{get_api_base_url_v2()}/forecast/list_ref_times",
        access_token=access_token,
        params=query,
    )

    return pd.DataFrame.from_dict(data)


def forecast_diff(
    comparison_type: str,
    series_ids: List[str],
    price_areas: List[str] = None,
    ref_time1: pd.Timestamp = None,
    ref_time2: pd.Timestamp = None,
    *,
    access_token: str = None,
):
    """
    Get the difference between two different forecasts.

    Args:
        comparison_type:
            The type of comparison to do to the two forecasts:
                forecast_update - how has the forecast for a specific time range been
                    updated. The two forecast reference times must be within ~60 hours
                    of each other. Otherwise, the output will be empty (because
                    the forecasts don't overlap).
                forecast_diff - compares the forecasts of any two dates to indicate how
                    different they are.
        series_ids:
            The ID of the data series to get (eg - 'MET/forecast_air_temperature_2m' or
            'MET/forecast_wind_speed_10m'). To specify multiple data series, include
            the series_ids parameter multiple times in the url.
            Retrieve the complete list of series using the list_ids resource.
        price_areas:
            The name of the price area(eg - 'NO2', 'NO5', or 'DE1-NO1'). To specify
            multiple price areas, include the price_areas parameter multiple times in
            the url.
            Retrieve the complete list of price areas available for a specified series
            ID using list_series_price_areas resource.
        ref_time1:
            Specify one of the two timestamps for when a forecast was created. Should
            be specified in ISO 8601 format (eg - '2021-11-29T06:00:00+00:00').
        ref_time2:
            Specify another of the two timestamps for when a forecast was
            created. Should be specified in ISO 8601 format
            (eg - '2021-11-29T06:00:00+00:00').
        access_token (str, optional):
            Access token for the API

    Returns:
        Dataframe with the diff of the two forecasts
    """
    assert comparison_type in ["forecast_update", "forecast_diff",], (
        f"Unknown comparison_type '{comparison_type}' "
        f"(should be either 'forecast_update' or 'forecast_diff')"
    )
    query = {
        "forecast_comparison_type": comparison_type,
        "series_ids": series_ids,
        "price_areas": price_areas,
        "ref_time1": ref_time1,
        "ref_time2": ref_time2,
    }

    data = make_api_request(
        f"{get_api_base_url_v2()}/forecast/diff",
        access_token=access_token,
        params=query,
    )

    return pd.DataFrame.from_dict(data)


def forecast_get(
    series_ids: List[str],
    price_areas: List[str] = None,
    start: str = None,
    end: str = None,
    *,
    access_token: str = None,
):
    """Get any number of forecast time series.

    This function can be used to fetch time series from the Archimedes Database.
    To see which series are available, use `list_ids()`.

    Example:
        >>> import archimedes
        >>> archimedes.get(
        >>>     series_ids=["MET/forecast_wind_speed_10m"],
        >>>     price_areas=["NO1", "NO2"],
        >>>     start="2022-04-20T06:00:00+00:00",
        >>>     end="2022-04-20T12:00:00+00:00",
        >>> )
        series_id                 MET/forecast_wind_speed_10m
        price_area                          NO1   NO2
        from_dt
        2020-06-20T04:00:00+00:00          1.30  1.30
        2020-06-20T05:00:00+00:00          1.35  1.35
        ...                                 ...   ...
        2020-06-28T03:00:00+00:00          0.53  0.53
        2020-06-28T04:00:00+00:00          0.55  0.55

    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (List[str], optional): The price areas to pick, all price areas if
                                           None. Defaults to None.
        start (str, optional): The first datetime to fetch (inclusive). Returns all if
                               None. Defaults to None.
        end (str, optional): The last datetime to fetch (exclusive). Returns all if
                             None. Defaults to None.
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with all the time series data

    Raises:
        HTTPError: If an HTTP error occurs when requesting the API.
        NoneAuth: If the user is unauthorized or if the authorization has expired.
    """

    if isinstance(series_ids, str):
        series_ids = [series_ids]

    if isinstance(price_areas, str):
        price_areas = [price_areas]

    start, end = get_start_date(start), get_end_date(end)

    queries = [
        {
            "start": start,
            "end": end,
            "series_ids": series_id,
            "price_areas": price_areas,
        }
        for series_id in series_ids
    ]

    base_url = get_api_base_url_v2()
    params_array = [
        dict(url=f"{base_url}/forecast/get", access_token=access_token, params=query)
        for query in queries
    ]
    data = execute_many(make_api_request, params_array)

    if len(data) == 0:
        df = pd.DataFrame(
            columns=["series_id", "from_dt", "ref_dt", "value", "price_area"]
        )
    else:
        df = pd.DataFrame.from_dict(data)
        df = df.sort_values(by=["from_dt", "ref_dt"])

    df["from_dt"] = pd.to_datetime(df["from_dt"])
    df["ref_dt"] = pd.to_datetime(df["ref_dt"])
    df = df.pivot_table(
        values="value",
        columns=["series_id", "price_area"],
        index=["from_dt", "ref_dt"],
        aggfunc="last",
    )
    return df


def forecast_get_by_ref_time_interval(  # pylint:disable=too-many-arguments
    series_id: str,
    price_area: str,
    start: pd.Timestamp = None,
    end: pd.Timestamp = None,
    forecast_interval: int = 24,
    day_ahead_hour: int = None,
    *,
    access_token: str = None,
):
    """
    Get a single forecast value for every hour. The value should be from the forecast
    that was generated at least forecast_interval hours prior.

    Args:
        series_id:
            The ID of the data series to get (eg - 'MET/forecast_air_temperature_2m' or
            'MET/forecast_wind_speed_10m'). Retrieve the complete list of series using
            the list_ids resource.
        price_area:
            The name of the price area(eg - 'NO2', 'SE3'). Retrieve the complete list
            of price areas available for a specified series ID using
            list_series_price_areas resource.
        start:
            The first datetime to fetch (inclusive). Returns all if not set. Should be
            specified in ISO 8601 format (eg - '2021-11-29T06:00:00+00:00')
        end:
            The last datetime to fetch (exclusive). Returns all if not set. Should be
            specified in ISO 8601 format (eg - '2021-11-30T06:00:00+00:00')
        forecast_interval:
            The number of hours earlier that the forecast must have been generated. In
            some cases, it could be older (if no forecast was generated at exactly that
            hour). NOTE - this is ignored if day_ahead_hour is set.
        day_ahead_hour:
            Used for day-ahead market. Indicates the hour of the day when the market
            closes (CET - Central European Time). Will return the forecast generated
            before this time on the previous day
            ('forecast_interval' will be set to 24). For example, if set to '12'
            (noon CET), the values shown for every hour of a specific day will be
            fetched from the most recent forecast generated before noon (most likely
            11am) on the previous day.
        access_token (str, optional): Access token for the API

    Returns:
        Dataframe with the forecasted values
    """
    query = {
        "series_id": series_id,
        "price_area": price_area,
        "start": start,
        "end": end,
        "forecast_interval": forecast_interval,
        "day_ahead_hour": day_ahead_hour,
    }

    data = make_api_request(
        f"{get_api_base_url_v2()}/forecast/get_by_ref_time_interval",
        access_token=access_token,
        params=query,
    )

    return pd.DataFrame.from_dict(data)



import pandas as pd

urls = {
    "iris": "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/iris.csv",
    "dots": "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/dots.csv",
    "anagrams": "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/anagrams.csv",
    "fmri": "https://raw.githubusercontent.com/mwaskom/seaborn-data/master/fmri.csv",
}


def list_data():
    """List the test datasets."""
    print("These are the test datasets:")
    for key, value in urls.items():
        print("\t - " + key)


def load_data(name: str) -> pd.DataFrame:
    """Load a test dataset.

    Data are here: https://github.com/mwaskom/seaborn-data

    Inputs:
        name: The name of the test data
    Returns:
        df: The test data
    """
    url = urls[name]
    df = pd.read_csv(url)
    return df



from typing import Dict, TypedDict

from pandas import Timestamp


class PredictionData(TypedDict):
    """
    Model for Prediction data
    """

    from_dt: Timestamp
    ref_dt: Timestamp
    run_dt: Timestamp
    data: Dict


class Prediction(PredictionData):
    """
    Model for Prediction
    """

    prediction_id: str



from pprint import pprint

import pandas as pd

_letters = list("abcdefghijklmnopqrstuvw")
_lst = list("abcdefghijklmnopqrstuvw")

for letter1 in _letters:
    for letter2 in _letters:
        _lst.append(letter1 + letter2)

for letter1 in _letters[0:8]:
    for letter2 in _letters:
        for letter3 in _letters:
            _lst.append(letter1 + letter2 + letter3)


REPLACEMENT_NAMES = _lst


def full_print(df: pd.DataFrame) -> None:
    """Prints the full DataFrame

    Example:
        >>> df = archimedes.load_data("fmri")
        >>> arhcimedes.full_print(df)
            subject  timepoint event    region    signal
        0        s13         18  stim  parietal -0.017552
        1         s5         14  stim  parietal -0.080883
        2        s12         18  stim  parietal -0.081033
        3        s11         18  stim  parietal -0.046134
        4        s10         18  stim  parietal -0.037970
        5         s9         18  stim  parietal -0.103513
        6         s8         18  stim  parietal -0.064408
        7         s7         18  stim  parietal -0.060526
        ...

    Args:
        df (pd.DataFrame): The dataframe that you want to print
    """
    pd.set_option("display.max_rows", None)
    print(df)
    # print(df.tail(1))
    pd.set_option("display.max_rows", 10)


def compact_print(
    df: pd.DataFrame, show_mapping: bool = False, all_rows: bool = False
) -> None:
    """Prints a compact version of the DataFrame

    Example:
        >>> df = archimedes.load_data("fmri")
        >>> arhcimedes.compact_print(df, True, False)
                a   b     c         d         e
        0     s13  18  stim  parietal -0.017552
        1      s5  14  stim  parietal -0.080883
        ...   ...  ..   ...       ...       ...
        1062  s11   7   cue   frontal -0.025367
        1063   s0   0   cue  parietal -0.006899
        ...
        [1064 rows x 5 columns]
        This is a compact version of the dataframe, with columns:
        {'a': 'subject', 'b': 'timepoint', 'c': 'event', 'd': 'region', 'e': 'signal'}

    Args:
        df (pd.DataFrame): The dataframe that you want to print
        show_mapping (bool, optional): Set to True to print the column name mapping. Defaults to False.
        all_rows (bool, optional): Set to True if you want to print all rows. Defaults to False.
    """
    df_ = df.copy()
    num_columns = len(df_.columns)
    actual_columns = df_.columns
    compact_columns = REPLACEMENT_NAMES[0:num_columns]
    mapping = dict(zip(compact_columns, actual_columns))
    df_.columns = compact_columns
    if all_rows:
        pd.set_option("display.max_rows", None)
        print(df_)
        pd.set_option("display.max_rows", 10)
    else:
        print(df_)
    if show_mapping:
        print("This is a compact version of the dataframe, with columns:")
        pprint(mapping)



from typing import List

import pandas as pd

from archimedes.data.common import get_api_base_url_v2
from archimedes.utils import get_end_date, get_start_date
from archimedes.utils.api_request import make_api_request
from archimedes.utils.split import get_queries_observation, get_queries_observation_json
from archimedes.utils.threaded_executor import execute_many


def get(  # pylint:disable=too-many-locals
    series_ids: List[str],
    price_areas: List[str] = None,
    start: str = None,
    end: str = None,
    *,
    access_token: str = None,
):
    """Get any number of time series.

    This function can be used to fetch time series from the Archimedes Database.
    To see which series are available, use `list_ids()`.

    Example:
        >>> import archimedes
        >>> archimedes.get(
        >>>     series_ids=["NP/AreaPrices"],
        >>>     price_areas=["NO1", "NO2"],
        >>>     start="2020-06-20T04:00:00+00:00",
        >>>     end="2020-06-28T04:00:00+00:00",
        >>> )
        series_id                 NP/AreaPrices
        price_area                          NO1   NO2
        from_dt
        2020-06-20T04:00:00+00:00          1.30  1.30
        2020-06-20T05:00:00+00:00          1.35  1.35
        ...                                 ...   ...
        2020-06-28T03:00:00+00:00          0.53  0.53
        2020-06-28T04:00:00+00:00          0.55  0.55

    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (List[str], optional): The price areas to pick, all price areas if
                                           None. Defaults to None.
        start (str, optional): The first datetime to fetch (inclusive). Returns all if
                               None. Defaults to None.
        end (str, optional): The last datetime to fetch (exclusive). Returns all if
                             None. Defaults to None.
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with all the time series data

    Raises:
        HTTPError: If an HTTP error occurs when requesting the API.
        NoneAuth: If the user is unauthorized or if the authorization has expired.
    """

    if not series_ids:
        return pd.DataFrame()

    if isinstance(series_ids, str):
        series_ids = [series_ids]

    if isinstance(price_areas, str):
        price_areas = [price_areas]

    start, end = get_start_date(start), get_end_date(end)

    observation_series_ids, observation_json_series_ids = _split_series_ids(
        series_ids, access_token
    )

    queries_by_type = {
        "observations": get_queries_observation(
            observation_series_ids, price_areas, start, end
        ),
        "observations_json": get_queries_observation_json(
            observation_json_series_ids, price_areas, start, end
        ),
    }

    def _make_observation_or_observation_json_request(request_type, params):
        request_url_path = (
            "data" if request_type == "observations" else "observation_json"
        )
        query_result = make_api_request(
            f"{get_api_base_url_v2()}/{request_url_path}/get",
            params=params,
            access_token=access_token,
        )
        return request_type, query_result

    def _merge_function(result, aggr):
        if aggr is None:
            aggr = {}
        observation_type, value = result
        if observation_type not in aggr.keys():
            aggr[observation_type] = []
        aggr[observation_type].extend(value)
        return aggr

    params_array = []
    for query_type, queries in queries_by_type.items():
        for query in queries:
            params_array.append(dict(request_type=query_type, params=query))

    api_results = execute_many(
        _make_observation_or_observation_json_request, params_array, _merge_function
    )

    observations = _process_observations(api_results.get("observations", []))
    observations_json = _process_observations_json(
        api_results.get("observations_json", [])
    )

    if observations_json.empty and observations.empty:
        return pd.DataFrame()

    if observations.empty:
        df = observations_json
    elif observations_json.empty:
        df = observations
    else:
        df = pd.merge(
            observations,
            observations_json,
            left_index=True,
            right_index=True,
            how="outer",
        )

    df.index = pd.to_datetime(df.index)

    return df


def get_latest(
    series_ids: List[str],
    price_areas: List[str] = None,
    *,
    access_token: str = None,
):
    """Get the most recent data for any number of time series.

    This function is similar to `get()`, but only fetches data from the past 48 hours,
    potentially including future hours as well (as in the case of Spot price data).

    @TODO: Add an argument `hours` that allows the 'lookback' period to be extended
    to an arbitrary number of hours.

    Example:
        >>> import archimedes
        >>> # Calling this function at 2020-03-15T10:15:00
        >>> archimedes.get_latest(
        >>>     series_ids=["NP/AreaPrices", "NP/ConsumptionImbalancePrices"],
        >>>     price_areas=["NO1"],
        >>> )
        series_id                 NP/AreaPrices  NP/ConsumptionImbalancePrices
        price_area                          NO1                            NO1
        from_dt
        2020-03-14T04:11:00+00:00          1.30                           1.30
        2020-03-14T05:12:00+00:00          1.35                           1.35
        ...                                 ...                            ...
        2020-03-15T22:00:00+00:00          0.53                            NaN
        2020-03-15T23:00:00+00:00          0.55                            NaN

    Args:
        series_ids (List[str]): The series ids to get.
        price_areas (List[str], optional): The price areas to pick, all price areas if
                                           None. Defaults to None.
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with the latest time series data

    Raises:
        HTTPError: If an HTTP error occurs when requesting the API.
        NoneAuth: If the user is unauthorized or if the authorization has expired.
    """
    now_dt = pd.Timestamp.now(tz="utc")
    start_dt = now_dt - pd.Timedelta(days=2)
    # +14 days should be enough in all cases now:
    end_dt = now_dt + pd.Timedelta(days=14)

    df = get(
        series_ids=series_ids,
        price_areas=price_areas,
        start=start_dt.isoformat(),
        end=end_dt.isoformat(),
        access_token=access_token,
    )

    return df


def _split_series_ids(series_ids, access_token=None):
    base_url = get_api_base_url_v2()
    all_series_ids_observations = make_api_request(
        f"{base_url}/data/list_ids", access_token=access_token
    )
    all_series_ids_observations_json = make_api_request(
        f"{base_url}/observation_json/list_ids",
        access_token=access_token,
    )
    series_ids_observations = [
        series_id
        for series_id in series_ids
        if series_id in all_series_ids_observations["series_id"]
    ]
    series_ids_observations_json = [
        series_id
        for series_id in series_ids
        if series_id in all_series_ids_observations_json["series_id"]
    ]

    # Add series ids that exist in none of the list_ids to both
    series_ids_none = [
        series_id
        for series_id in series_ids
        if series_id not in series_ids_observations
        and series_id not in series_ids_observations_json
    ]
    series_ids_observations.extend(series_ids_none)
    series_ids_observations_json.extend(series_ids_none)

    return series_ids_observations, series_ids_observations_json


def _process_observations(observations):
    observations = pd.DataFrame(observations)
    if not observations.empty:
        observations = observations.sort_values(by=["from_dt", "version"]).pivot_table(
            values="value",
            columns=["series_id", "price_area"],
            index="from_dt",
            aggfunc="last",
        )
    return observations


def _process_observations_json(observations):
    observations_json = pd.DataFrame(observations)
    if not observations_json.empty:
        observations_json[
            ["series_id1", "series_id2", "price_area"]
        ] = observations_json["series_id"].str.split("/", 2, expand=True)
        observations_json["series_id"] = (
            observations_json["series_id1"] + "/" + observations_json["series_id2"]
        )
        observations_json = observations_json.drop(["series_id1", "series_id2"], axis=1)
        is_intraday_trades = observations_json["series_id"] == "NP/IntradayTrades"

        def agg_func(items):
            return list(items)

        observation_data_intraday = observations_json[is_intraday_trades]
        if len(observation_data_intraday):
            observation_data_intraday = observation_data_intraday.sort_values(
                by=["from_dt"]
            )
            observation_data_intraday = observation_data_intraday.pivot_table(
                values="value",
                columns=["series_id", "price_area"],
                index="from_dt",
                aggfunc=agg_func,
            )

        observations_json = observations_json[~is_intraday_trades]
        if len(observations_json):
            observations_json = observations_json.sort_values(
                by=["from_dt", "version"]
            ).pivot_table(
                values="value",
                columns=["series_id", "price_area"],
                index="from_dt",
                aggfunc="last",
            )

        if observation_data_intraday.empty and observations_json.empty:
            observations_json = pd.DataFrame()
        elif observation_data_intraday.empty:
            pass
        elif observations_json.empty:
            observations_json = observation_data_intraday
        else:
            observations_json = observations_json.merge(
                observation_data_intraday,
                left_index=True,
                right_index=True,
                how="outer",
            )
    return observations_json



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['archimedes',
 'archimedes.data',
 'archimedes.data.api',
 'archimedes.testdata',
 'archimedes.utils',
 'archimedes.utils.split']

package_data = \
{'': ['*'], 'archimedes.testdata': ['datasets/*']}

install_requires = \
['iteration-utilities>=0.11.0,<0.12.0',
 'msal-extensions>=1.0.0,<2.0.0',
 'msal>=1.10.0,<2.0.0',
 'pandas>=1.0.5',
 'requests>=2.27.0',
 'retry>=0.9.2,<0.10.0']

setup_kwargs = {
    'name': 'archimedes-python-client',
    'version': '4.6.2',
    'description': 'The Python library for Archimedes',
    'long_description': None,
    'author': 'Optimeering AS',
    'author_email': 'dev@optimeering.com',
    'maintainer': None,
    'maintainer_email': None,
    'url': None,
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'python_requires': '>=3.8,<3.11',
}


setup(**setup_kwargs)



import json
from typing import Dict, List

import pandas as pd

from archimedes.data.common import get_api_base_url_v2
from archimedes.data.types import PredictionData
from archimedes.utils.api_request import make_api_request
from archimedes.utils.threaded_executor import execute_many


def get_predictions(  # pylint:disable=too-many-arguments
    series_ids: List[str] = None,
    price_areas: List[str] = None,
    start: str = None,
    end: str = None,
    ref_dt_start: str = None,
    ref_dt_end: str = None,
    *,
    access_token: str = None,
) -> pd.DataFrame:
    """Get any number of predictions

    This function can be used to fetch predictions from the Archimedes Database.

    Unlike `archimedes.get`, this will return a list, not a dataframe.

    Example:
        >>> import archimedes
        >>> archimedes.get_predictions(
        >>>    series_ids=["PX/rk-naive"],
        >>>    price_areas=["NO1"],
        >>>    start="2020"
        >>> )
        >>> [...]

    Args:
        series_ids (List[str], optional): The series ids to get.
        price_areas (List[str], optional): The price areas to get the data for.
        start (str, optional):
            The first datetime to fetch (inclusive). Returns all if None.
            Defaults to None.
        end (str, optional):
            The last datetime to fetch (exclusive). Returns all if None.
            Defaults to None.
        ref_dt_start (pd.Timestamp, optional):
            The earliest ref_dt to fetch (inclusive). Defaults to None.
        ref_dt_end (pd.Timestamp, optional):
            The latest ref_dt to fetch (exclusive). Defaults to None.
        access_token (str, optional): None - access token for the API

    Returns:
        DataFrame with all the prediction data
    """
    if isinstance(series_ids, str):
        series_ids = [series_ids]

    if isinstance(price_areas, str):
        price_areas = [price_areas]

    query = {}

    if start is not None:
        query["start"] = pd.to_datetime(start, utc=True)

    if end is not None:
        query["end"] = pd.to_datetime(end, utc=True)

    if ref_dt_start is not None:
        query["ref_dt_start"] = pd.to_datetime(ref_dt_start, utc=True)

    if ref_dt_end is not None:
        query["ref_dt_end"] = pd.to_datetime(ref_dt_end, utc=True)

    queries = [query]
    if series_ids is not None:
        queries = [
            {"series_ids": series_id, "price_areas": price_areas, **query}
            for query in queries
            for series_id in series_ids
        ]

    def _merge_function(result, aggr):
        if aggr is None:
            aggr = []
        for item in result:
            if "json_data" in item:
                json_data = item.get("json_data")
                item.update(json_data)
                del item["json_data"]
            aggr.append(item)
        return aggr

    params_array = [
        dict(
            url=f"{get_api_base_url_v2()}/data/get_predictions",
            access_token=access_token,
            params=query,
        )
        for query in queries
    ]

    data = execute_many(make_api_request, params_array, _merge_function)

    df = pd.DataFrame.from_dict(data)

    date_fields = ["from_dt", "run_dt", "ref_dt"]

    for date_field in date_fields:
        if date_field in df:
            df[date_field] = pd.to_datetime(df[date_field], utc=True)

    df = df.fillna("")

    return df


def get_predictions_ref_dts(prediction_id: str = None, *, access_token: str = None):
    """Get which ref_dts are available.

    ref_dt == prediction_build_dt
    Users views in the database.

    Args:
        prediction_id (str): The series id to get the reference dts for. If None, get
                             ref_dts for all prediction_ids.
        access_token (str, optional): Access token for the API

    Returns:
        DataFrame with all ref_dts
    """
    query = {}

    if prediction_id:
        query["prediction_id"] = prediction_id

    data = make_api_request(
        f"{get_api_base_url_v2()}/data/get_predictions_ref_dts",
        access_token=access_token,
        params=query,
    )

    return pd.DataFrame.from_dict(data)


def store_prediction(
    prediction_id: str,
    from_dt: pd.Timestamp,
    ref_dt: pd.Timestamp,
    run_dt: pd.Timestamp,
    data: Dict,
    *,
    access_token: str = None,
):
    """Store a prediction

    Example:
        >>> import archimedes
        >>> import pandas as pd
        >>> from dateutil.tz import gettz
        >>> tz_oslo = gettz("Europe/Oslo")
        >>> d = {
        >>>     "direction": "D",
        >>>     "probability": 0.8632089971077396,
        >>>     "hours_ahead": 1,
        >>>     "price_area": "NO1"
        >>> }
        >>> archimedes.store_prediction(
        >>>     prediction_id="test-prediction-id",
        >>>     from_dt=pd.Timestamp("2021-04-11 23:47:16.854775807", tz=tz_oslo),
        >>>     ref_dt=pd.Timestamp("2021-04-10 23:00:00.000000000", tz=tz_oslo),
        >>>     run_dt=pd.Timestamp.now(tz=tz_oslo),
        >>>     data=d
        >>> )
        True
    """
    payload = {
        "prediction_id": prediction_id,
        "from_dt": from_dt.isoformat(),
        "ref_dt": ref_dt.isoformat(),
        "run_dt": run_dt.isoformat(),
        "data": data,
    }

    ret = make_api_request(
        f"{get_api_base_url_v2()}/data/store_prediction",
        method="POST",
        data=json.dumps(payload),
        headers={"Content-Type": "application/json"},
        access_token=access_token,
    )

    return ret.get("success", False)


def store_predictions(
    prediction_id: str,
    prediction_data: List[PredictionData],
    *,
    access_token: str = None,
):
    """Store a prediction

    Example:
        >>> import archimedes
        >>> import pandas as pd
        >>> from dateutil.tz import gettz
        >>> tz_oslo = gettz("Europe/Oslo")
        >>> pid = "test-prediction-id"
        >>> from_dt = pd.Timestamp("2021-04-11 23:47:16.854775807", tz=tz_oslo)
        >>> ref_dt = pd.Timestamp("2021-04-10 23:00:00.000000000", tz=tz_oslo)
        >>> run_dt = pd.Timestamp.now(tz=tz_oslo)
        >>> data = {
        >>>     "direction": "D",
        >>>     "probability": 0.8632089971077396,
        >>>     "hours_ahead": 1,
        >>>     "price_area": "NO1"
        >>> }
        >>> p_data = [
        >>>    {'from_dt': from_dt, 'ref_dt': ref_dt, 'run_dt': run_dt, 'data': data }
        >>> ]
        >>> archimedes.store_predictions(
        >>>     prediction_id=pid,
        >>>     prediction_data=p_data
        >>> )
        True
    """

    predictions = [
        {
            "from_dt": p["from_dt"].isoformat(),
            "ref_dt": p["ref_dt"].isoformat(),
            "run_dt": p["run_dt"].isoformat(),
            "data": p["data"],
        }
        for p in prediction_data
    ]

    payload = {
        "prediction_id": prediction_id,
        "data": predictions,
    }

    ret = make_api_request(
        f"{get_api_base_url_v2()}/data/store_predictions",
        method="POST",
        data=json.dumps(payload),
        headers={"Content-Type": "application/json"},
        access_token=access_token,
    )

    return ret.get("success", False)
